from typing import Any, List, Mapping, Optional

from langchain.callbacks.manager import CallbackManagerForLLMRun
from langchain.embeddings.base import Embeddings
from langchain.llms.base import LLM
from langchain.vectorstores import Qdrant
from pydantic import BaseModel
import requests

from ..api.api_http import headers
from ..api.api_request import provision_req
from ..api.api_substation import SubstationAPI
from ..configuration import config
from ..util import unwrap
from ..vector import vector_store


class SeaplaneEmbeddingFunction(BaseModel, Embeddings):
    url = config.substation_embed_endpoint

    def _embed(self, query: str) -> List[float]:
        req = provision_req(config._token_api)

        result = unwrap(
            req(
                lambda access_token: requests.post(
                    self.url,
                    headers=headers(access_token),
                    json={"query": query},
                )
            )
        )

        return [float(i) for i in result["embeddings"]]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        embeddings = []
        for text in texts:
            response = self._embed(f"Represent the document for Retrieval: {text}")
            embeddings.append(response)
        return embeddings

    def embed_query(self, query: str) -> List[float]:
        response = self._embed(f"Represent the Science sentence: {query}")
        return response


seaplane_embeddings = SeaplaneEmbeddingFunction()


def langchain_vectorstore(index_name: str, embeddings: Embeddings = seaplane_embeddings) -> Qdrant:
    vectorstore = Qdrant(
        client=vector_store._get_client(),
        collection_name=index_name,
        embeddings=embeddings,
    )

    return vectorstore


class SeaplaneLLM(LLM):
    max_output_length: Optional[int] = 3000
    model_specific_prompt: Optional[bool] = True

    @property
    def _llm_type(self) -> str:
        return "custom"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> str:
        if stop is not None:
            raise ValueError("stop kwargs are not permitted.")
        substation_api = SubstationAPI(config)
        result = substation_api.predict(prompt, self.max_output_length, self.model_specific_prompt)
        return str(result["choices"][0]["text"]["generated_text"])

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {"model": "MPT-30B", "provider": "Seaplane"}



from base64 import urlsafe_b64decode, urlsafe_b64encode


def base64url_encode_from_bytes(data: bytes) -> str:
    return urlsafe_b64encode(data).rstrip(b"=").decode()


def base64url_decode_to_bytes(data: str) -> bytes:
    padding = b"=" * (4 - (len(data) % 4))
    return urlsafe_b64decode(data.encode() + padding)



import secrets
import string
from typing import Any, Callable, List, Optional

from ..logging import log
from .task import Task, TaskEvent


def random_id() -> str:
    alphabet = string.ascii_letters + string.digits
    random_string = "".join(secrets.choice(alphabet) for i in range(10))
    return random_string


class AppEvent:
    def __init__(self, app_id: str, input: Any) -> None:
        self.id = random_id()
        self.app_id = app_id
        self.status = "in_progress"
        self.input: Any = input
        self.output: Optional[Any] = None
        self.tasks: List[TaskEvent] = []
        self.error: Optional[Any] = None

    def add_task_event(self, event: TaskEvent) -> None:
        for i, cp in enumerate(self.tasks):
            if cp.id == event.id:
                self.tasks[i] = event
                return
        self.tasks.append(event)

    def set_output(self, output: Any) -> None:
        self.output = output
        self.status = "completed"

    def set_error(self, error: Any) -> None:
        self.error = error
        self.status = "error"


class App:
    def __init__(
        self,
        func: Callable[[Any], Any],
        path: str,
        method: str,
        id: str,
        parameters: Optional[List[str]],
    ) -> None:
        self.func = func
        self.path = path
        self.method = method
        self.id = id
        self.parameters = parameters
        self.tasks: List[Task] = []
        self.events: List[AppEvent] = []
        self.return_source = None

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.func(*args, *kwargs)

    def add_task(self, task: Task) -> None:
        for i, cp in enumerate(self.tasks):
            if cp.id == task.id:
                self.tasks[i] = task
                return
        self.tasks.append(task)

    def print(self) -> None:
        log.info(f"id: {self.id}, path: {self.path}, method: {self.method}")
